/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Copy IPv4 TCP entry data to traceroute
int
copy_tcp4_to_traceroute (SPSData *data) {

  int i;

  // Ethernet Header
  memcpy (data->ethhdr[9].dst_mac, data->ethhdr[0].dst_mac, 6 * sizeof (uint8_t));  // Destination link-layer (MAC) address (48 bits)
  memcpy (data->ifname[9], data->ifname[0], TMP_STRINGLEN * sizeof (char));  // Source Interface Name
  data->ifmtu[9] = data->ifmtu[0];  // MTU
  memcpy (data->ethhdr[9].src_mac, data->ethhdr[0].src_mac, 6 * sizeof (uint8_t));
  data->ethhdr[9].type_code = data->ethhdr[0].type_code;  // Ethernet type code (16 bits)

  // IPv4 Header
  data->ip4hdr[9].ip_hl = data->ip4hdr[0].ip_hl;  // IPv4 header length (4 bits)
  data->ip4hdr[9].ip_v = data->ip4hdr[0].ip_v;  // Internet Protocol version (4 bits)
  data->ip4hdr[9].ip_tos = data->ip4hdr[0].ip_tos;  // Type of service (8 bits)
  data->ip4hdr[9].ip_len = data->ip4hdr[0].ip_len;  // Total length of datagram (16 bits)
  data->ip4hdr[9].ip_id = data->ip4hdr[0].ip_id;  // ID sequence number (16 bits)
  data->ip4hdr[9].ip_off = data->ip4hdr[0].ip_off;  // Flags, and Fragmentation offset (3, 13 bits)
  data->ip4hdr[9].ip_ttl = data->ip4hdr[0].ip_ttl;  // Time-to-Live (8 bits)
  data->ip4hdr[9].ip_p = IPPROTO_IPV6;  // Transport layer protocol (8 bits)
  memcpy (&(data->ip4hdr[9].ip_src), &(data->ip4hdr[0].ip_src), 4 * sizeof (char));  // Source IPv4 address (32 bits)
  memcpy (&(data->ip4hdr[9].ip_dst), &(data->ip4hdr[0].ip_dst), 4 * sizeof (char));  // Destination IPv4 address (32 bits)
  data->ip4hdr[9].ip_sum = data->ip4hdr[0].ip_sum;  // IPv4 header checksum (16 bits)

  // IP Options
  data->ip_nopt[9] = data->ip_nopt[0];  // Number of IP header options
  for (i=0; i<data->ip_nopt[9]; i++) {
    data->ip_optlen[9][i] = data->ip_optlen[0][i];  // Length of option i
    memcpy (data->ip_options[9][i], data->ip_options[0][i], data->ip_optlen[10][i] * sizeof (uint8_t));  // Copy option i
  }
  data->ip_optlenbuf[9] = data->ip_optlenbuf[0];  // Length of option in buffer

  // TCP Header
  data->ran_tcp4_tr_sourceport = data->ran_tcp4_sourceport;  // Randomize TCP source port
  data->tcphdr[9].th_sport = data->tcphdr[0].th_sport;  // Source port number (16 bits)
  data->tcphdr[9].th_dport = data->tcphdr[0].th_dport;  // Destination port number (16 bits)
  data->tcphdr[9].th_seq = data->tcphdr[0].th_seq;  // Sequence number (32 bits)
  data->tcphdr[9].th_ack = data->tcphdr[0].th_ack;  // Acknowledgement number (32 bits)
  data->tcphdr[9].th_x2 = data->tcphdr[0].th_x2;  // Reserved (4 bits)
  data->tcphdr[9].th_off = data->tcphdr[0].th_off;  // Data offset (4 bits)
  data->tcphdr[9].th_flags = data->tcphdr[0].th_flags;  // Flags (8 bits)
  data->tcphdr[9].th_win = data->tcphdr[0].th_win;  // Window size (16 bits)
  data->tcphdr[9].th_urp = data->tcphdr[0].th_urp;  // Urgent pointer (16 bits)
  data->tcphdr[9].th_sum = data->tcphdr[0].th_sum;  // TCP checksum (16 bits)

  // TCP Options
  data->tcp_nopt[9] = data->ip_nopt[0];  // Number of TCP header options
  for (i=0; i<data->tcp_nopt[9]; i++) {
    data->tcp_optlen[9][i] = data->tcp_optlen[0][i];  // Length of option i
    memcpy (data->tcp_options[9][i], data->tcp_options[0][i], data->tcp_optlen[9][i] * sizeof (uint8_t));  // Copy option i
  }
  data->tcp_opt_totlen[9] = data->tcp_opt_totlen[0];  // Total length of TCP options
  data->tcp_optlenbuf[9] = data->tcp_optlenbuf[0];  // Length of option in buffer

  // TCP Data
  data->payloadlen[9] = data->payloadlen[0];  // Length of TCP payload
  memcpy (data->payload[9], data->payload[0], data->payloadlen[9] * sizeof (uint8_t));  // Copy TCP payload data

  // Populate entries in traceroute.ui window.
  tcp4_tr_show (data);

  return (EXIT_SUCCESS);
}

// Copy IPv4 ICMP entry data to traceroute
int
copy_icmp4_to_traceroute (SPSData *data) {

  int i;

  // Ethernet Header
  memcpy (data->ethhdr[10].dst_mac, data->ethhdr[1].dst_mac, 6 * sizeof (uint8_t));  // Destination link-layer (MAC) address (48 bits)
  memcpy (data->ifname[10], data->ifname[1], TMP_STRINGLEN * sizeof (char));  // Source Interface Name
  data->ifmtu[10] = data->ifmtu[1];  // MTU
  memcpy (data->ethhdr[10].src_mac, data->ethhdr[1].src_mac, 6 * sizeof (uint8_t));  // Source link-layer (MAC) address (48 bits)
  data->ethhdr[10].type_code = data->ethhdr[1].type_code;  // Ethernet type code (16 bits)

  // IPv4 Header
  data->ip4hdr[10].ip_hl = data->ip4hdr[1].ip_hl;  // IPv4 header length (4 bits)
  data->ip4hdr[10].ip_v = data->ip4hdr[1].ip_v;  // Internet Protocol version (4 bits)
  data->ip4hdr[10].ip_tos = data->ip4hdr[1].ip_tos;  // Type of service (8 bits)
  data->ip4hdr[10].ip_len = data->ip4hdr[1].ip_len;  // Total length of datagram (16 bits)
  data->ip4hdr[10].ip_id = data->ip4hdr[1].ip_id;  // ID sequence number (16 bits)
  data->ip4hdr[10].ip_off = data->ip4hdr[1].ip_off;  // Flags, and Fragmentation offset (3, 13 bits)
  data->ip4hdr[10].ip_ttl = data->ip4hdr[1].ip_ttl;  // Time-to-Live (8 bits)
  data->ip4hdr[10].ip_p = IPPROTO_IPV6;  // Transport layer protocol (8 bits)
  memcpy (&(data->ip4hdr[10].ip_src), &(data->ip4hdr[1].ip_src), 4 * sizeof (char));  // Source IPv4 address (32 bits)
  memcpy (&(data->ip4hdr[10].ip_dst), &(data->ip4hdr[1].ip_dst), 4 * sizeof (char));  // Destination IPv4 address (32 bits)
  data->ip4hdr[10].ip_sum = data->ip4hdr[1].ip_sum;  // IPv4 header checksum (16 bits)

  // IP Options
  data->ip_nopt[10] = data->ip_nopt[1];  // Number of IP header options
  for (i=0; i<data->ip_nopt[10]; i++) {
    data->ip_optlen[10][i] = data->ip_optlen[1][i];  // Length of option i
    memcpy (data->ip_options[10][i], data->ip_options[1][i], data->ip_optlen[10][i] * sizeof (uint8_t));  // Copy option i
  }
  data->ip_optlenbuf[10] = data->ip_optlenbuf[1];  // Length of option in buffer

  // ICMP Header
  data->icmp4hdr[10].icmp_type = data->icmp4hdr[1].icmp_type;  // Message Type (8 bits)
  data->icmp4hdr[10].icmp_code = data->icmp4hdr[1].icmp_code;  // Message Code (8 bits)
  data->icmp4hdr[10].icmp_id = data->icmp4hdr[1].icmp_id;  // Identifier (16 bits)
  data->icmp4hdr[10].icmp_seq = data->icmp4hdr[1].icmp_seq;  // Sequence Number (16 bits)
  data->icmp4hdr[10].icmp_cksum = data->icmp4hdr[1].icmp_cksum;  // ICMP header checksum (16 bits)

  // ICMP Data
  data->payloadlen[10] = data->payloadlen[1];  // Length of ICMP payload data
  memcpy (data->payload[10], data->payload[1], data->payloadlen[10] * sizeof (uint8_t));  // Copy ICMP payload data

  // Populate entries in traceroute.ui window.
  icmp4_tr_show (data);

  return (EXIT_SUCCESS);
}

// Copy IPv4 UDP entry data to traceroute
int
copy_udp4_to_traceroute (SPSData *data) {

  int i;

  // Ethernet Header
  memcpy (data->ethhdr[11].dst_mac, data->ethhdr[2].dst_mac, 6 * sizeof (uint8_t));  // Destination link-layer (MAC) address (48 bits)
  memcpy (data->ifname[11], data->ifname[2], TMP_STRINGLEN * sizeof (char));  // Source Interface Name
  data->ifmtu[11] = data->ifmtu[2];  // MTU
  memcpy (data->ethhdr[11].src_mac, data->ethhdr[2].src_mac, 6 * sizeof (uint8_t));  // Source link-layer (MAC) address (48 bits)
  data->ethhdr[11].type_code = data->ethhdr[2].type_code;  // Ethernet type code (16 bits)

  // IPv4 Header
  data->ip4hdr[11].ip_hl = data->ip4hdr[2].ip_hl;  // IPv4 header length (4 bits)
  data->ip4hdr[11].ip_v = data->ip4hdr[2].ip_v;  // Internet Protocol version (4 bits)
  data->ip4hdr[11].ip_tos = data->ip4hdr[2].ip_tos;  // Type of service (8 bits)
  data->ip4hdr[11].ip_len = data->ip4hdr[2].ip_len;  // Total length of datagram (16 bits)
  data->ip4hdr[11].ip_id = data->ip4hdr[2].ip_id;  // ID sequence number (16 bits)
  data->ip4hdr[11].ip_off = data->ip4hdr[2].ip_off;  // Flags, and Fragmentation offset (3, 13 bits)
  data->ip4hdr[11].ip_ttl = data->ip4hdr[2].ip_ttl;  // Time-to-Live (8 bits)
  data->ip4hdr[11].ip_p = data->ip4hdr[2].ip_p;  // Transport layer protocol (8 bits)
  memcpy (&(data->ip4hdr[11].ip_src), &(data->ip4hdr[2].ip_src), 4 * sizeof (char));  // Source IPv4 address (32 bits)
  memcpy (&(data->ip4hdr[11].ip_dst), &(data->ip4hdr[2].ip_dst), 4 * sizeof (char));  // Destination IPv4 address (32 bits)
  data->ip4hdr[11].ip_sum = data->ip4hdr[2].ip_sum;  // // IPv4 header checksum (16 bits)

  // IP Options
  data->ip_nopt[11] = data->ip_nopt[2];  // Number of IP header options
  for (i=0; i<data->ip_nopt[11]; i++) {
    data->ip_optlen[11][i] = data->ip_optlen[2][i];  // Length of option i
    memcpy (data->ip_options[11][i], data->ip_options[2][i], data->ip_optlen[11][i] * sizeof (uint8_t));  // Copy option i
  }
  data->ip_optlenbuf[11] = data->ip_optlenbuf[2];  // Length of option in buffer

  // UDP Header
  data->ran_udp4_tr_sourceport = data->ran_udp4_sourceport;  // Randomize UDP source port
  data->udphdr[11].uh_sport = data->udphdr[2].uh_sport;  // Source port number (16 bits)
  data->udphdr[11].uh_dport = data->udphdr[2].uh_dport;  // Destination port number (16 bits)
  data->udphdr[11].uh_ulen = data->udphdr[2].uh_ulen;  // Length of UDP datagram (16 bits)
  data->udphdr[11].uh_sum = data->udphdr[2].uh_sum;  // UDP checksum (16 bits)

  // UDP Data
  data->payloadlen[11] = data->payloadlen[2];  // Length of UDP payload data
  memcpy (data->payload[11], data->payload[2], data->payloadlen[11] * sizeof (uint8_t));  // Copy UDP payload data

  // Populate entries in traceroute.ui window.
  udp4_tr_show (data);

  return (EXIT_SUCCESS);
}

// Copy IPv6 TCP entry data to traceroute
int
copy_tcp6_to_traceroute (SPSData *data) {

  int i;

  // Ethernet Header
  memcpy (data->ethhdr[12].dst_mac, data->ethhdr[3].dst_mac, 6 * sizeof (uint8_t));  // Destination link-layer (MAC) address (48 bits)
  memcpy (data->ifname[12], data->ifname[3], TMP_STRINGLEN * sizeof (char));  // Source Interface Name
  data->ifmtu[12] = data->ifmtu[3];  // MTU
  memcpy (data->ethhdr[12].src_mac, data->ethhdr[3].src_mac, 6 * sizeof (uint8_t));
  data->ethhdr[12].type_code = data->ethhdr[3].type_code;  // Ethernet type code (16 bits)

  // IPv6 Header
  data->ip6hdr[12].ip6_flow = data->ip6hdr[3].ip6_flow;  // IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr[12].ip6_plen = data->ip6hdr[3].ip6_plen;  // Payload length (16 bits)
  data->ip6hdr[12].ip6_nxt = data->ip6hdr[3].ip6_nxt;  // Next header (8 bits)
  data->ip6hdr[12].ip6_hops = data->ip6hdr[3].ip6_hops;  // Hop limit (8 bits)
  memcpy (&(data->ip6hdr[12].ip6_src), &(data->ip6hdr[3].ip6_src), 16 * sizeof (uint8_t));  // Source IPv6 address (128 bits)
  memcpy (&(data->ip6hdr[12].ip6_dst), &(data->ip6hdr[3].ip6_dst), 16 * sizeof (uint8_t));  // Destination IPv6 address (128 bits)

  // TCP Header
  data->ran_tcp6_tr_sourceport = data->ran_tcp6_sourceport;  // Randomize TCP source port
  data->tcphdr[12].th_sport = data->tcphdr[3].th_sport;  // Source port number (16 bits)
  data->tcphdr[12].th_dport = data->tcphdr[3].th_dport;  // Destination port number (16 bits)
  data->tcphdr[12].th_seq = data->tcphdr[3].th_seq;  // Sequence number (32 bits)
  data->tcphdr[12].th_ack = data->tcphdr[3].th_ack;  // Acknowledgement number (32 bits)
  data->tcphdr[12].th_x2 = data->tcphdr[3].th_x2;  // Reserved (4 bits)
  data->tcphdr[12].th_off = data->tcphdr[3].th_off;  // Data offset (4 bits)
  data->tcphdr[12].th_flags = data->tcphdr[3].th_flags;  // Flags (8 bits)
  data->tcphdr[12].th_win = data->tcphdr[3].th_win;  // Window size (16 bits)
  data->tcphdr[12].th_urp = data->tcphdr[3].th_urp;  // Urgent pointer (16 bits)
  data->tcphdr[12].th_sum = data->tcphdr[3].th_sum;  // TCP checksum (16 bits)

  // TCP Options
  data->tcp_nopt[12] = data->tcp_nopt[3];  // Number of TCP options
  for (i=0; i<data->tcp_nopt[12]; i++) {
    data->tcp_optlen[12][i] = data->tcp_optlen[3][i];  // Length of option i
    memcpy (data->tcp_options[12][i], data->tcp_options[3][i], data->tcp_optlen[12][i] * sizeof (uint8_t));  // Copy option i
  }
  data->tcp_opt_totlen[12] = data->tcp_opt_totlen[3];  // Total length of TCP options
  data->tcp_optlenbuf[12] = data->tcp_optlenbuf[3];  // Length of option in buffer

  // TCP Data
  data->payloadlen[12] = data->payloadlen[3];  // Length of TCP payload data
  memcpy (data->payload[12], data->payload[3], data->payloadlen[12] * sizeof (uint8_t));  // Copy TCP payload data

  // Hop-by-Hop Options
  data->hbh_hdr_flag[12][0] = data->hbh_hdr_flag[3][0];  // Hop-by-Hop header flag
  data->hbh_hdr_flag[12][1] = data->hbh_hdr_flag[3][1];  // Hop-by-Hop header flag
  data->hbh_nopt[12][0] = data->hbh_nopt[3][0];  // Number of hop-by-hop options
  data->hbh_nopt[12][1] = data->hbh_nopt[3][1];  // Number of hop-by-hop options
  for (i=0; i<data->hbh_nopt[12][0]; i++) {
    data->hbh_optlen[12][0][i] = data->hbh_optlen[3][0][i];  // Length of option i
    memcpy (data->hbh_options[12][0][i], data->hbh_options[3][0][i], data->hbh_optlen[12][0][i] * sizeof (uint8_t));  // Copy option i
    data->hbh_x[12][0][i] = data->hbh_x[3][0][i];  // Option alignment parameter x (of xN + y)
    data->hbh_y[12][0][i] = data->hbh_y[3][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->hbh_nopt[12][1]; i++) {
    data->hbh_optlen[12][1][i] = data->hbh_optlen[3][1][i];  // Length of option i
    memcpy (data->hbh_options[12][1][i], data->hbh_options[3][1][i], data->hbh_optlen[12][1][i] * sizeof (uint8_t));  // Copy option i
    data->hbh_x[12][1][i] = data->hbh_x[3][1][i];  // Option alignment parameter x (of xN + y)
    data->hbh_y[12][1][i] = data->hbh_y[3][1][i];  // Option alignment parameter y (of xN + y)
  }
  data->hbh_optpadlen[12][0] = data->hbh_optpadlen[3][0];  // Hop-by-hop header options padding length (includes alignment padding)
  data->hbh_optpadlen[12][1] = data->hbh_optpadlen[3][1];  // Hop-by-hop header options padding length (includes alignment padding)
  data->hbh_opt_totlen[12][0] = data->hbh_opt_totlen[3][0];  // Total length of hop-by-hop options
  data->hbh_opt_totlen[12][1] = data->hbh_opt_totlen[3][1];  // Total length of hop-by-hop options
  data->hophdr[12][0].hdr_len = data->hophdr[3][0].hdr_len;  // Hop-by-hop header length
  data->hophdr[12][1].hdr_len = data->hophdr[3][1].hdr_len;  // Hop-by-hop header length

  // Destination Options
  data->dstf_hdr_flag[12][0] = data->dstf_hdr_flag[3][0];  // Flag for first destination header
  data->dstf_hdr_flag[12][1] = data->dstf_hdr_flag[3][1];  // Flag for first destination header
  data->dstf_nopt[12][0] = data->dstf_nopt[3][0];  // Number of first destination header options
  data->dstf_nopt[12][1] = data->dstf_nopt[3][1];  // Number of first destination header options
  data->dstl_hdr_flag[12][0] = data->dstl_hdr_flag[3][0];  // Flag for last destination header
  data->dstl_hdr_flag[12][1] = data->dstl_hdr_flag[3][1];  // Flag for last destination header
  data->dstl_nopt[12][0] = data->dstl_nopt[3][0];  // Number of last destination header options
  data->dstl_nopt[12][1] = data->dstl_nopt[3][1];  // Number of last destination header options
  for (i=0; i<data->dstf_nopt[12][0]; i++) {
    data->dstf_optlen[12][0][i] = data->dstf_optlen[3][0][i];  // Length of first destination option i
    memcpy (data->dstf_options[12][0][i], data->dstf_options[3][0][i], data->dstf_optlen[12][0][i] * sizeof (uint8_t));  // Copy option i
    data->dstf_x[12][0][i] = data->dstf_x[3][0][i];  // Option alignment parameter x (of xN + y)
    data->dstf_y[12][0][i] = data->dstf_y[3][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstf_nopt[12][1]; i++) {
    data->dstf_optlen[12][1][i] = data->dstf_optlen[3][1][i];  // Length of first destination option i
    memcpy (data->dstf_options[12][1][i], data->dstf_options[3][1][i], data->dstf_optlen[12][1][i] * sizeof (uint8_t));  // Copy option i
    data->dstf_x[12][1][i] = data->dstf_x[3][1][i];  // Option alignment parameter x (of xN + y)
    data->dstf_y[12][1][i] = data->dstf_y[3][1][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstl_nopt[12][0]; i++) {
    data->dstl_optlen[12][0][i] = data->dstl_optlen[3][0][i];  // Length of last destination option i
    memcpy (data->dstl_options[12][0][i], data->dstl_options[3][0][i], data->dstl_optlen[12][0][i] * sizeof (uint8_t));  // Copy option i
    data->dstl_x[12][0][i] = data->dstl_x[3][0][i];  // Option alignment parameter x (of xN + y)
    data->dstl_y[12][0][i] = data->dstl_y[3][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstl_nopt[12][1]; i++) {
    data->dstl_optlen[12][1][i] = data->dstl_optlen[3][1][i];  // Length of last destination option i
    memcpy (data->dstl_options[12][1][i], data->dstl_options[3][1][i], data->dstl_optlen[12][1][i] * sizeof (uint8_t));  // Copy option i
    data->dstl_x[12][1][i] = data->dstl_x[3][1][i];  // Option alignment parameter x (of xN + y)
    data->dstl_y[12][1][i] = data->dstl_y[3][1][i];  // Option alignment parameter y (of xN + y)
  }

  // Routing Header
  data->route_hdr_flag[12][0] = data->route_hdr_flag[3][0];  // Flag for routing header
  data->route_hdr_flag[12][1] = data->route_hdr_flag[3][1];  // Flag for routing header
  data->route_datlen[12][0] = data->route_datlen[3][0];  // Length of routing header data
  data->route_datlen[12][1] = data->route_datlen[3][1];  // Length of routing header data
  data->routehdr[12][0].hdr_len = data->routehdr[3][0].hdr_len;  // Length of routing header (units of 8 bytes, excluding first 8 bytes)
  data->routehdr[12][1].hdr_len = data->routehdr[3][1].hdr_len;  // Length of routing header (units of 8 bytes, excluding first 8 bytes)
  data->routehdr[12][0].routing_type = data->routehdr[3][0].routing_type;  // Routing Type (8 bits)
  data->routehdr[12][1].routing_type = data->routehdr[3][1].routing_type;  // Routing Type (8 bits)
  data->routehdr[12][0].segs_left = data->routehdr[3][0].segs_left;  // Segments Left (8 bits)
  data->routehdr[12][1].segs_left = data->routehdr[3][1].segs_left;  // Segments Left (8 bits)
  memcpy (data->route_data[12][0], data->route_data[3][0], MAX_ROUTEDATA * sizeof (uint8_t));  // Copy routing data
  memcpy (data->route_data[12][1], data->route_data[3][1], MAX_ROUTEDATA * sizeof (uint8_t));  // Copy routing data

  // Authentication Header
  data->auth_hdr_flag[12] = data->auth_hdr_flag[3];  // Flag for authentication header
  data->auth_len[12] = data->auth_len[3];  // Length of authentication header
  data->authhdr[12].pay_len = data->authhdr[3].pay_len;  // Authentication header payload length (8 bits)
  data->authhdr[12].reserved = data->authhdr[3].reserved;  // Reserved (16 bits)
  data->authhdr[12].spi = data->authhdr[3].spi;  // Security Parameters Index (SPI) (32 bits)
  data->authhdr[12].seq = data->authhdr[3].seq;  // Sequence Number (32 bits)
  memcpy (data->auth_data[12], data->auth_data[3], MAX_AUTHICVLEN * sizeof (uint8_t));  // Copy Integrity Check Value (ICV)

  // Encapsulating security payload (ESP) Header
  data->esp_hdr_flag[12] = data->esp_hdr_flag[3];  // Flag for ESP header
  data->esp_auth_len[12] = data->esp_auth_len[3];  // Length of authentication
  data->esphdr[12].spi = data->esphdr[3].spi;  // Security Parameters Index (SPI) (32 bits)
  data->esphdr[12].seq = data->esphdr[3].seq;  // Sequence Number (32 bits)
  memcpy (data->esp_auth_data[12], data->esp_auth_data[3], MAX_ESPICVLEN * sizeof (uint8_t));  // Copy ESP authentication data
  data->esptail[12].pad_len = data->esptail[3].pad_len;  // ESP Trailer Padding Length (8 bits)
  memcpy (data->esp_pad[12], data->esp_pad[3], data->esptail[12].pad_len * sizeof (uint8_t));  // Copy padding
  data->esptail[12].nxt_hdr = data->esptail[3].nxt_hdr;  // ESP trailer Next Header field (8 bits)

  // Populate entries in traceroute.ui window.
  tcp6_tr_show (data);

  return (EXIT_SUCCESS);
}

// Copy IPv6 ICMP entry data to traceroute
int
copy_icmp6_to_traceroute (SPSData *data) {

  int i;

  // Ethernet Header
  memcpy (data->ethhdr[13].dst_mac, data->ethhdr[4].dst_mac, 6 * sizeof (uint8_t));  // Destination link-layer (MAC) address (48 bits)
  memcpy (data->ifname[13], data->ifname[4], TMP_STRINGLEN * sizeof (char));  // Source Interface Name
  data->ifmtu[13] = data->ifmtu[4];  // MTU
  memcpy (data->ethhdr[13].src_mac, data->ethhdr[4].src_mac, 6 * sizeof (uint8_t));
  data->ethhdr[13].type_code = data->ethhdr[4].type_code;  // Ethernet type code (16 bits)

  // IPv6 Header
  data->ip6hdr[13].ip6_flow = data->ip6hdr[4].ip6_flow;  // IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr[13].ip6_plen = data->ip6hdr[4].ip6_plen;  // Payload length (16 bits)
  data->ip6hdr[13].ip6_nxt = data->ip6hdr[4].ip6_nxt;  // Next header (8 bits)
  data->ip6hdr[13].ip6_hops = data->ip6hdr[4].ip6_hops;  // Hop limit (8 bits)
  memcpy (&(data->ip6hdr[13].ip6_src), &(data->ip6hdr[4].ip6_src), 16 * sizeof (uint8_t));  // Source IPv6 address (128 bits)
  memcpy (&(data->ip6hdr[13].ip6_dst), &(data->ip6hdr[4].ip6_dst), 16 * sizeof (uint8_t));  // Destination IPv6 address (128 bits)

  // ICMP Header
  data->icmp4hdr[13].icmp_type = data->icmp4hdr[4].icmp_type;  // Message Type (8 bits)
  data->icmp4hdr[13].icmp_code = data->icmp4hdr[4].icmp_code;  // Message Code (8 bits)
  data->icmp4hdr[13].icmp_id = data->icmp4hdr[4].icmp_id;  // Identifier (16 bits)
  data->icmp4hdr[13].icmp_seq = data->icmp4hdr[4].icmp_seq;  // Sequence Number (16 bits)
  data->icmp4hdr[13].icmp_cksum = data->icmp4hdr[4].icmp_cksum;  // ICMP header checksum (16 bits)

  // ICMP Data
  data->payloadlen[13] = data->payloadlen[4];  // Length of ICMP payload data
  memcpy (data->payload[13], data->payload[4], data->payloadlen[13] * sizeof (uint8_t));  // Copy ICMP payload data

  // Hop-by-Hop Options
  data->hbh_hdr_flag[13][0] = data->hbh_hdr_flag[4][0];  // Hop-by-Hop header flag
  data->hbh_hdr_flag[13][1] = data->hbh_hdr_flag[4][1];  // Hop-by-Hop header flag
  data->hbh_nopt[13][0] = data->hbh_nopt[4][0];  // Number of hop-by-hop options
  data->hbh_nopt[13][1] = data->hbh_nopt[4][1];  // Number of hop-by-hop options
  for (i=0; i<data->hbh_nopt[13][0]; i++) {
    data->hbh_optlen[13][0][i] = data->hbh_optlen[4][0][i];  // Length of option i
    memcpy (data->hbh_options[13][0][i], data->hbh_options[4][0][i], data->hbh_optlen[13][0][i] * sizeof (uint8_t));  // Copy option i
    data->hbh_x[13][0][i] = data->hbh_x[4][0][i];  // Option alignment parameter x (of xN + y)
    data->hbh_y[13][0][i] = data->hbh_y[4][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->hbh_nopt[13][1]; i++) {
    data->hbh_optlen[13][1][i] = data->hbh_optlen[4][1][i];  // Length of option i
    memcpy (data->hbh_options[13][1][i], data->hbh_options[4][1][i], data->hbh_optlen[13][1][i] * sizeof (uint8_t));  // Copy option i
    data->hbh_x[13][1][i] = data->hbh_x[4][1][i];  // Option alignment parameter x (of xN + y)
    data->hbh_y[13][1][i] = data->hbh_y[4][1][i];  // Option alignment parameter y (of xN + y)
  }
  data->hbh_optpadlen[13][0] = data->hbh_optpadlen[4][0];  // Hop-by-hop header options padding length (includes alignment padding)
  data->hbh_optpadlen[13][1] = data->hbh_optpadlen[4][1];  // Hop-by-hop header options padding length (includes alignment padding)
  data->hbh_opt_totlen[13][0] = data->hbh_opt_totlen[4][0];  // Total length of hop-by-hop options
  data->hbh_opt_totlen[13][1] = data->hbh_opt_totlen[4][1];  // Total length of hop-by-hop options
  data->hophdr[13][0].hdr_len = data->hophdr[4][0].hdr_len;  // Hop-by-hop header length
  data->hophdr[13][1].hdr_len = data->hophdr[4][1].hdr_len;  // Hop-by-hop header length

  // Destination Options
  data->dstf_hdr_flag[13][0] = data->dstf_hdr_flag[4][0];  // Flag for first destination header
  data->dstf_hdr_flag[13][1] = data->dstf_hdr_flag[4][1];  // Flag for first destination header
  data->dstf_nopt[13][0] = data->dstf_nopt[4][0];  // Number of first destination header options
  data->dstf_nopt[13][1] = data->dstf_nopt[4][1];  // Number of first destination header options
  data->dstl_hdr_flag[13][0] = data->dstl_hdr_flag[4][0];  // Flag for last destination header
  data->dstl_hdr_flag[13][1] = data->dstl_hdr_flag[4][1];  // Flag for last destination header
  data->dstl_nopt[13][0] = data->dstl_nopt[4][0];  // Number of last destination header options
  data->dstl_nopt[13][1] = data->dstl_nopt[4][1];  // Number of last destination header options
  for (i=0; i<data->dstf_nopt[13][0]; i++) {
    data->dstf_optlen[13][0][i] = data->dstf_optlen[4][0][i];  // Length of first destination option i
    memcpy (data->dstf_options[13][0][i], data->dstf_options[4][0][i], data->dstf_optlen[13][0][i] * sizeof (uint8_t));  // Copy option i  
    data->dstf_x[13][0][i] = data->dstf_x[4][0][i];  // Option alignment parameter x (of xN + y)
    data->dstf_y[13][0][i] = data->dstf_y[4][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstf_nopt[13][1]; i++) {
    data->dstf_optlen[13][1][i] = data->dstf_optlen[4][1][i];  // Length of first destination option i
    memcpy (data->dstf_options[13][1][i], data->dstf_options[4][1][i], data->dstf_optlen[13][1][i] * sizeof (uint8_t));  // Copy option i  
    data->dstf_x[13][1][i] = data->dstf_x[4][1][i];  // Option alignment parameter x (of xN + y)
    data->dstf_y[13][1][i] = data->dstf_y[4][1][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstl_nopt[13][0]; i++) {
    data->dstl_optlen[13][0][i] = data->dstl_optlen[4][0][i];  // Length of last destination option i
    memcpy (data->dstl_options[13][0][i], data->dstl_options[4][0][i], data->dstl_optlen[13][0][i] * sizeof (uint8_t));  // Copy option i  
    data->dstl_x[13][0][i] = data->dstl_x[4][0][i];  // Option alignment parameter x (of xN + y)
    data->dstl_y[13][0][i] = data->dstl_y[4][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstl_nopt[13][1]; i++) {
    data->dstl_optlen[13][1][i] = data->dstl_optlen[4][1][i];  // Length of last destination option i
    memcpy (data->dstl_options[13][1][i], data->dstl_options[4][1][i], data->dstl_optlen[13][1][i] * sizeof (uint8_t));  // Copy option i  
    data->dstl_x[13][1][i] = data->dstl_x[4][1][i];  // Option alignment parameter x (of xN + y)
    data->dstl_y[13][1][i] = data->dstl_y[4][1][i];  // Option alignment parameter y (of xN + y)
  }
  
  // Routing Header
  data->route_hdr_flag[13][0] = data->route_hdr_flag[4][0];  // Flag for routing header
  data->route_hdr_flag[13][1] = data->route_hdr_flag[4][1];  // Flag for routing header
  data->route_datlen[13][0] = data->route_datlen[4][0];  // Length of routing header data
  data->route_datlen[13][1] = data->route_datlen[4][1];  // Length of routing header data
  data->routehdr[13][0].hdr_len = data->routehdr[4][0].hdr_len;  // Length of routing header (units of 8 bytes, excluding first 8 bytes) 
  data->routehdr[13][1].hdr_len = data->routehdr[4][1].hdr_len;  // Length of routing header (units of 8 bytes, excluding first 8 bytes) 
  data->routehdr[13][0].routing_type = data->routehdr[4][0].routing_type;  // Routing Type (8 bits)
  data->routehdr[13][1].routing_type = data->routehdr[4][1].routing_type;  // Routing Type (8 bits)
  data->routehdr[13][0].segs_left = data->routehdr[4][0].segs_left;  // Segments Left (8 bits)
  data->routehdr[13][1].segs_left = data->routehdr[4][1].segs_left;  // Segments Left (8 bits)
  memcpy (data->route_data[13][0], data->route_data[4][0], MAX_ROUTEDATA * sizeof (uint8_t));  // Copy routing data
  memcpy (data->route_data[13][1], data->route_data[4][1], MAX_ROUTEDATA * sizeof (uint8_t));  // Copy routing data

  // Authentication Header
  data->auth_hdr_flag[13] = data->auth_hdr_flag[4];  // Flag for authentication header
  data->auth_len[13] = data->auth_len[4];  // Length of authentication header
  data->authhdr[13].pay_len = data->authhdr[4].pay_len;  // Authentication header payload length (8 bits)
  data->authhdr[13].reserved = data->authhdr[4].reserved;  // Reserved (16 bits)
  data->authhdr[13].spi = data->authhdr[4].spi;  // Security Parameters Index (SPI) (32 bits)
  data->authhdr[13].seq = data->authhdr[4].seq;  // Sequence Number (32 bits)
  memcpy (data->auth_data[13], data->auth_data[4], MAX_AUTHICVLEN * sizeof (uint8_t));  // Copy Integrity Check Value (ICV)

  // Encapsulating security payload (ESP) Header
  data->esp_hdr_flag[13] = data->esp_hdr_flag[4];  // Flag for ESP header
  data->esp_auth_len[13] = data->esp_auth_len[4];  // Length of authentication
  data->esphdr[13].spi = data->esphdr[4].spi;  // Security Parameters Index (SPI) (32 bits)
  data->esphdr[13].seq = data->esphdr[4].seq;  // Sequence Number (32 bits)
  memcpy (data->esp_auth_data[13], data->esp_auth_data[4], MAX_ESPICVLEN * sizeof (uint8_t));  // Copy ESP authentication data
  data->esptail[13].pad_len = data->esptail[4].pad_len;  // ESP Trailer Padding Length (8 bits)
  memcpy (data->esp_pad[13], data->esp_pad[4], data->esptail[13].pad_len * sizeof (uint8_t));  // Copy padding
  data->esptail[13].nxt_hdr = data->esptail[4].nxt_hdr;  // ESP trailer Next Header field (8 bits)

  // Populate entries in traceroute.ui window.
  icmp6_tr_show (data);

  return (EXIT_SUCCESS);
}

// Copy IPv6 UDP entry data to traceroute
int
copy_udp6_to_traceroute (SPSData *data) {

  int i;

  // Ethernet Header
  memcpy (data->ethhdr[14].dst_mac, data->ethhdr[5].dst_mac, 6 * sizeof (uint8_t));  // Destination link-layer (MAC) address (48 bits)
  memcpy (data->ifname[14], data->ifname[5], TMP_STRINGLEN * sizeof (char));  // Source Interface Name
  data->ifmtu[14] = data->ifmtu[5];  // MTU
  memcpy (data->ethhdr[14].src_mac, data->ethhdr[5].src_mac, 6 * sizeof (uint8_t));
  data->ethhdr[14].type_code = data->ethhdr[5].type_code;  // Ethernet type code (16 bits)

  // IPv6 Header
  data->ip6hdr[14].ip6_flow = data->ip6hdr[5].ip6_flow;  // IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr[14].ip6_plen = data->ip6hdr[5].ip6_plen;  // Payload length (16 bits)
  data->ip6hdr[14].ip6_nxt = data->ip6hdr[5].ip6_nxt;  // Next header (8 bits)
  data->ip6hdr[14].ip6_hops = data->ip6hdr[5].ip6_hops;  // Hop limit (8 bits)
  memcpy (&(data->ip6hdr[14].ip6_src), &(data->ip6hdr[5].ip6_src), 16 * sizeof (uint8_t));  // Source IPv6 address (128 bits)
  memcpy (&(data->ip6hdr[14].ip6_dst), &(data->ip6hdr[5].ip6_dst), 16 * sizeof (uint8_t));  // Destination IPv6 address (128 bits)

  // UDP Header
  data->ran_udp6_tr_sourceport = data->ran_udp6_sourceport;  // Randomize UDP source port
  data->udphdr[14].uh_sport = data->udphdr[5].uh_sport;  // Source port number (16 bits)
  data->udphdr[14].uh_dport = data->udphdr[5].uh_dport;  // Destination port number (16 bits)
  data->udphdr[14].uh_ulen = data->udphdr[5].uh_ulen;  // Length of UDP datagram (16 bits)
  data->udphdr[14].uh_sum = data->udphdr[5].uh_sum;  // UDP checksum (16 bits)

  // UDP Data
  data->payloadlen[14] = data->payloadlen[5];  // Length of UDP payload data 
  memcpy (data->payload[14], data->payload[5], data->payloadlen[14] * sizeof (uint8_t));  // Copy UDP payload data

  // Hop-by-Hop Options
  data->hbh_hdr_flag[14][0] = data->hbh_hdr_flag[5][0];  // Hop-by-Hop header flag
  data->hbh_hdr_flag[14][1] = data->hbh_hdr_flag[5][1];  // Hop-by-Hop header flag
  data->hbh_nopt[14][0] = data->hbh_nopt[5][0];  // Number of hop-by-hop options
  data->hbh_nopt[14][1] = data->hbh_nopt[5][1];  // Number of hop-by-hop options
  for (i=0; i<data->hbh_nopt[14][0]; i++) {
    data->hbh_optlen[14][0][i] = data->hbh_optlen[5][0][i];  // Length of option i
    memcpy (data->hbh_options[14][0][i], data->hbh_options[5][0][i], data->hbh_optlen[14][0][i] * sizeof (uint8_t));  // Copy option i
    data->hbh_x[14][0][i] = data->hbh_x[5][0][i];  // Option alignment parameter x (of xN + y)
    data->hbh_y[14][0][i] = data->hbh_y[5][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->hbh_nopt[14][1]; i++) {
    data->hbh_optlen[14][1][i] = data->hbh_optlen[5][1][i];  // Length of option i
    memcpy (data->hbh_options[14][1][i], data->hbh_options[5][1][i], data->hbh_optlen[14][1][i] * sizeof (uint8_t));  // Copy option i
    data->hbh_x[14][1][i] = data->hbh_x[5][1][i];  // Option alignment parameter x (of xN + y)
    data->hbh_y[14][1][i] = data->hbh_y[5][1][i];  // Option alignment parameter y (of xN + y)
  }
  data->hbh_optpadlen[14][0] = data->hbh_optpadlen[5][0];  // Hop-by-hop header options padding length (includes alignment padding)
  data->hbh_optpadlen[14][1] = data->hbh_optpadlen[5][1];  // Hop-by-hop header options padding length (includes alignment padding)
  data->hbh_opt_totlen[14][0] = data->hbh_opt_totlen[5][0];  // Total length of hop-by-hop options
  data->hbh_opt_totlen[14][1] = data->hbh_opt_totlen[5][1];  // Total length of hop-by-hop options
  data->hophdr[14][0].hdr_len = data->hophdr[5][0].hdr_len;  // Hop-by-hop header length
  data->hophdr[14][1].hdr_len = data->hophdr[5][1].hdr_len;  // Hop-by-hop header length

  // Destination Options
  data->dstf_hdr_flag[14][0] = data->dstf_hdr_flag[5][0];  // Flag for first destination header
  data->dstf_hdr_flag[14][1] = data->dstf_hdr_flag[5][1];  // Flag for first destination header
  data->dstf_nopt[14][0] = data->dstf_nopt[5][0];  // Number of first destination header options
  data->dstf_nopt[14][1] = data->dstf_nopt[5][1];  // Number of first destination header options
  data->dstl_hdr_flag[14][0] = data->dstl_hdr_flag[5][0];  // Flag for last destination header
  data->dstl_hdr_flag[14][1] = data->dstl_hdr_flag[5][1];  // Flag for last destination header
  data->dstl_nopt[14][0] = data->dstl_nopt[5][0];  // Number of last destination header options
  data->dstl_nopt[14][1] = data->dstl_nopt[5][1];  // Number of last destination header options
  for (i=0; i<data->dstf_nopt[14][0]; i++) {
    data->dstf_optlen[14][0][i] = data->dstf_optlen[5][0][i];  // Length of first destination option i
    memcpy (data->dstf_options[14][0][i], data->dstf_options[5][0][i], data->dstf_optlen[14][0][i] * sizeof (uint8_t));  // Copy option i  
    data->dstf_x[14][0][i] = data->dstf_x[5][0][i];  // Option alignment parameter x (of xN + y)
    data->dstf_y[14][0][i] = data->dstf_y[5][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstf_nopt[14][1]; i++) {
    data->dstf_optlen[14][1][i] = data->dstf_optlen[5][1][i];  // Length of first destination option i
    memcpy (data->dstf_options[14][1][i], data->dstf_options[5][1][i], data->dstf_optlen[14][1][i] * sizeof (uint8_t));  // Copy option i  
    data->dstf_x[14][1][i] = data->dstf_x[5][1][i];  // Option alignment parameter x (of xN + y)
    data->dstf_y[14][1][i] = data->dstf_y[5][1][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstl_nopt[14][0]; i++) {
    data->dstl_optlen[14][0][i] = data->dstl_optlen[5][0][i];  // Length of last destination option i
    memcpy (data->dstl_options[14][0][i], data->dstl_options[5][0][i], data->dstl_optlen[14][0][i] * sizeof (uint8_t));  // Copy option i  
    data->dstl_x[14][0][i] = data->dstl_x[5][0][i];  // Option alignment parameter x (of xN + y)
    data->dstl_y[14][0][i] = data->dstl_y[5][0][i];  // Option alignment parameter y (of xN + y)
  }
  for (i=0; i<data->dstl_nopt[14][1]; i++) {
    data->dstl_optlen[14][1][i] = data->dstl_optlen[5][1][i];  // Length of last destination option i
    memcpy (data->dstl_options[14][1][i], data->dstl_options[5][1][i], data->dstl_optlen[14][1][i] * sizeof (uint8_t));  // Copy option i  
    data->dstl_x[14][1][i] = data->dstl_x[5][1][i];  // Option alignment parameter x (of xN + y)
    data->dstl_y[14][1][i] = data->dstl_y[5][1][i];  // Option alignment parameter y (of xN + y)
  }

  // Routing Header
  data->route_hdr_flag[14][0] = data->route_hdr_flag[5][0];  // Flag for routing header
  data->route_hdr_flag[14][1] = data->route_hdr_flag[5][1];  // Flag for routing header
  data->route_datlen[14][0] = data->route_datlen[5][0];  // Length of routing header data
  data->route_datlen[14][1] = data->route_datlen[5][1];  // Length of routing header data
  data->routehdr[14][0].hdr_len = data->routehdr[5][0].hdr_len;  // Length of routing header (units of 8 bytes, excluding first 8 bytes) 
  data->routehdr[14][1].hdr_len = data->routehdr[5][1].hdr_len;  // Length of routing header (units of 8 bytes, excluding first 8 bytes) 
  data->routehdr[14][0].routing_type = data->routehdr[5][0].routing_type;  // Routing Type (8 bits)
  data->routehdr[14][1].routing_type = data->routehdr[5][1].routing_type;  // Routing Type (8 bits)
  data->routehdr[14][0].segs_left = data->routehdr[5][0].segs_left;  // Segments Left (8 bits)
  data->routehdr[14][1].segs_left = data->routehdr[5][1].segs_left;  // Segments Left (8 bits)
  memcpy (data->route_data[14][0], data->route_data[5][0], MAX_ROUTEDATA * sizeof (uint8_t));  // Copy routing data
  memcpy (data->route_data[14][1], data->route_data[5][1], MAX_ROUTEDATA * sizeof (uint8_t));  // Copy routing data

  // Authentication Header
  data->auth_hdr_flag[14] = data->auth_hdr_flag[5];  // Flag for authentication header
  data->auth_len[14] = data->auth_len[5];  // Length of authentication header
  data->authhdr[14].pay_len = data->authhdr[5].pay_len;  // Authentication header payload length (8 bits)
  data->authhdr[14].reserved = data->authhdr[5].reserved;  // Reserved (16 bits)
  data->authhdr[14].spi = data->authhdr[5].spi;  // Security Parameters Index (SPI) (32 bits)
  data->authhdr[14].seq = data->authhdr[5].seq;  // Sequence Number (32 bits)
  memcpy (data->auth_data[14], data->auth_data[5], MAX_AUTHICVLEN * sizeof (uint8_t));  // Copy Integrity Check Value (ICV)

  // Encapsulating security payload (ESP) Header
  data->esp_hdr_flag[14] = data->esp_hdr_flag[5];  // Flag for ESP header
  data->esp_auth_len[14] = data->esp_auth_len[5];  // Length of authentication
  data->esphdr[14].spi = data->esphdr[5].spi;  // Security Parameters Index (SPI) (32 bits)
  data->esphdr[14].seq = data->esphdr[5].seq;  // Sequence Number (32 bits)
  memcpy (data->esp_auth_data[14], data->esp_auth_data[5], MAX_ESPICVLEN * sizeof (uint8_t));  // Copy ESP authentication data
  data->esptail[14].pad_len = data->esptail[5].pad_len;  // ESP Trailer Padding Length (8 bits)
  memcpy (data->esp_pad[14], data->esp_pad[5], data->esptail[14].pad_len * sizeof (uint8_t));  // Copy padding
  data->esptail[14].nxt_hdr = data->esptail[5].nxt_hdr;  // ESP trailer Next Header field (8 bits)

  // Populate entries in traceroute.ui window.
  udp6_tr_show (data);

  return (EXIT_SUCCESS);
}
