/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Obtain current local host date and time.
// Return in character string.
char *
date_and_time (char *text, int maxlen) {

  time_t rawtime;
  struct tm *timeinfo;
  char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  char *temp;

  // Allocate memory for array.
  temp = allocate_strmem (TEXT_STRINGLEN);

  // Grab system time.
  time (&rawtime);

  // Express time as per local timezone, and as broken-down time.
  timeinfo = localtime (&rawtime);

  // Add day.
  memset (text, 0, maxlen * sizeof (char));
  sprintf (text, "%i ", timeinfo->tm_mday);

  // Append month.
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, "%.3s ", mon_name[timeinfo->tm_mon]);
  strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);

  // Append year.
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, "%i ", timeinfo->tm_year + 1900);
  strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);

  // Append time.
  memset (temp, 0, TEXT_STRINGLEN * sizeof (char));
  sprintf (temp, "%.2d:%.2d:%.2d  ", timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
  strncat (text, temp, TEXT_STRINGLEN - strnlen (text, TEXT_STRINGLEN - 1) - 1);

  // Free allocated memory.
  free (temp);

  return (text);
}
