/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Destination (original/inner) extension header editing window
int
dst_win (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If destination window is already open, return with no action.
  if (data->dst_flag[0]) {
    return (EXIT_FAILURE);
  } else {
    data->dst_flag[0] = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "destination.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->dst_window = GTK_WIDGET (gtk_builder_get_object (builder, "dst_window"));

  data->dst_radiobutton61 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton61"));  // Decimal data entry
  data->dst_radiobutton62 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton62"));  // Hexadecimal data entry

  data->dst_button143 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_checkbutton143"));  // Toggle: attach / detach a "first" destination header
  data->dst_textview48 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview48"));  // Status: attached / detached
  data->dst_button144 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_checkbutton144"));  // Toggle: attach / detach a "last" destination header
  data->dst_textview49 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview49"));  // Status: attached / detached
  data->dst_entry439 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry439"));  // Destination header length (first)
  data->dst_entry440 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry440"));  // Destination header length (last)
  data->dst_entry315 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry315"));  // Destination option data
  data->dst_entry156 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry156"));  // Option alignment parameter x (of xN + y)
  data->dst_entry157 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry157"));  // Option alignment parameter y (of xN + y)
  data->dst_radiobutton63 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton63"));  // Designate as first destination header
  data->dst_radiobutton64 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton64"));  // Designate as last destination header
  data->dst_button104 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button104"));  // Insert Option After...
  data->dst_entry431 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry431"));  // Option # to insert after
  data->dst_button105 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button105"));  // Append option
  data->dst_textview25 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview25"));  // Number of options in first destination header
  data->dst_textview27 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview27"));  // Number of options in last destination header
  data->dst_button106 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button106"));  // View destination options (first)
  data->dst_button109 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button109"));  // View destination options (last)
  data->dst_entry432 = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry432"));  // Option number to remove
  data->dst_button107 = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button107"));  // Remove option

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->dst_window);

  // Populate/set widgets in destination header editing window.
  dst_show (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Populate dst_window with contents of selected destination header.
int
dst_show (int type, SPSData *data) {

  char *text;
  GtkTextBuffer *textbuffer25, *textbuffer27, *textbuffer48, *textbuffer49;

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);

  // Flag indicating a "first" destination header
  textbuffer48 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview48));
  if (data->dstf_hdr_flag[data->exthdr_type][0]) {
    gtk_text_buffer_set_text (textbuffer48, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer48, "Not Attached", -1);
  }

  // Flag indicating a "last" destination header
  textbuffer49 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview49));
  if (data->dstl_hdr_flag[data->exthdr_type][0]) {
    gtk_text_buffer_set_text (textbuffer49, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer49, "Not Attached", -1);
  }

  // Decimal or hexadecimal data entry
  // destination.ui will always start with hexdecimal checked.
  data->dec_hex_dstopt[0] = 1;

  // Designate data entry to be for last destination header.
  // destination.ui will always start with last destination header checked.
  data->dst_fl_flag[0] = 1;

  // Destination extension header (first) length (header + alignments + options + padding)
  sprintf (text, "%" PRIu8, (uint8_t) data->dstfhdr[type][0].hdr_len);
  gtk_entry_set_text (GTK_ENTRY (data->dst_entry439), text);

  // Destination extension header (last) length (header + alignments + options + padding)
  sprintf (text, "%" PRIu8, (uint8_t) data->dstlhdr[type][0].hdr_len);
  gtk_entry_set_text (GTK_ENTRY (data->dst_entry440), text);

  // Show number of options in first destination header.
  sprintf (text, "%i", data->dstf_nopt[type][0]);
  textbuffer25 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25));
  gtk_text_buffer_set_text (textbuffer25, text, -1);

  memset (text, 0, TEXT_STRINGLEN * sizeof (char));

  // Show number of options in last destination header.
  sprintf (text, "%i", data->dstl_nopt[type][0]);
  textbuffer27 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27));
  gtk_text_buffer_set_text (textbuffer27, text, -1);

  // Free allocated memory.
  free (text);

  return (EXIT_SUCCESS);
}

// Destination (outer) extension header editing window
int
dst_win_outer (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If destination window is already open, return with no action.
  if (data->dst_flag[1]) {
    return (EXIT_FAILURE);
  } else {
    data->dst_flag[1] = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "destination-outer.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->dst_window_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_window_outer"));

  data->dst_radiobutton61_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton61_outer"));  // Decimal data entry
  data->dst_radiobutton62_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton62_outer"));  // Hexadecimal data entry

  data->dst_button143_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_checkbutton143_outer"));  // Toggle: attach / detach a "first" destination header
  data->dst_textview48_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview48_outer"));  // Status: attached / detached
  data->dst_button144_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_checkbutton144_outer"));  // Toggle: attach / detach a "last" destination header
  data->dst_textview49_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview49_outer"));  // Status: attached / detached
  data->dst_entry439_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry439_outer"));  // Destination header length (first)
  data->dst_entry440_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry440_outer"));  // Destination header length (last)
  data->dst_entry315_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry315_outer"));  // Destination option data
  data->dst_entry156_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry156_outer"));  // Option alignment parameter x (of xN + y)
  data->dst_entry157_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry157_outer"));  // Option alignment parameter y (of xN + y)
  data->dst_radiobutton63_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton63_outer"));  // Designate as first destination header
  data->dst_radiobutton64_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "dst_radiobutton64_outer"));  // Designate as last destination header
  data->dst_button104_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button104_outer"));  // Insert Option After...
  data->dst_entry431_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry431_outer"));  // Option # to insert after
  data->dst_button105_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button105_outer"));  // Append option
  data->dst_textview25_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview25_outer"));  // Number of options in first destination header
  data->dst_textview27_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_textview27_outer"));  // Number of options in last destination header
  data->dst_button106_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button106_outer"));  // View destination options (first)
  data->dst_button109_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button109_outer"));  // View destination options (last)
  data->dst_entry432_outer = GTK_WIDGET (gtk_builder_get_object (builder, "dst_entry432_outer"));  // Option number to remove
  data->dst_button107_outer = GTK_BUTTON (gtk_builder_get_object (builder, "dst_button107_outer"));  // Remove option

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->dst_window_outer);

  // Populate/set widgets in destination header editing window.
  dst_show_outer (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Populate dst_window_outer with contents of selected destination header.
int
dst_show_outer (int type, SPSData *data) {

  char *text;
  GtkTextBuffer *textbuffer25_outer, *textbuffer27_outer, *textbuffer48_outer, *textbuffer49_outer;

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);

  // Flag indicating a "first" destination header
  textbuffer48_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview48_outer));
  if (data->dstf_hdr_flag[data->exthdr_type][1]) {
    gtk_text_buffer_set_text (textbuffer48_outer, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer48_outer, "Not Attached", -1);
  }

  // Flag indicating a "last" destination header
  textbuffer49_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview49_outer));
  if (data->dstl_hdr_flag[data->exthdr_type][1]) {
    gtk_text_buffer_set_text (textbuffer49_outer, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer49_outer, "Not Attached", -1);
  }

  // Decimal or hexadecimal data entry
  // destination.ui will always start with hexdecimal checked.
  data->dec_hex_dstopt[1] = 1;

  // Designate data entry to be for last destination header.
  // destination.ui will always start with last destination header checked.
  data->dst_fl_flag[1] = 1;

  // Destination extension header (first) length (header + alignments + options + padding)
  sprintf (text, "%" PRIu8, (uint8_t) data->dstfhdr[type][1].hdr_len);
  gtk_entry_set_text (GTK_ENTRY (data->dst_entry439_outer), text);

  // Destination extension header (last) length (header + alignments + options + padding)
  sprintf (text, "%" PRIu8, (uint8_t) data->dstlhdr[type][1].hdr_len);
  gtk_entry_set_text (GTK_ENTRY (data->dst_entry440_outer), text);

  // Show number of options in first destination header.
  sprintf (text, "%i", data->dstf_nopt[type][1]);
  textbuffer25_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25_outer));
  gtk_text_buffer_set_text (textbuffer25_outer, text, -1);

  memset (text, 0, TEXT_STRINGLEN * sizeof (char));

  // Show number of options in last destination header.
  sprintf (text, "%i", data->dstl_nopt[type][1]);
  textbuffer27_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27_outer));
  gtk_text_buffer_set_text (textbuffer27_outer, text, -1);

  // Free allocated memory.
  free (text);

  return (EXIT_SUCCESS);
}

// Insert, remove, or append a destination option.
int
dst_option (int io, int mode, int opt_number, SPSData *data) {

  // Values for io:
  // 0 = original/inner destination header
  // 1 = outer destination header

  // Values for mode:
  // 0 = insert
  // 1 = remove
  // 2 = append

  const char *entryx_text, *entryy_text;
  int i, j, maxoptions, maxoptlen, hdrlen;
  int *nopt, *optlen, *x, *y, *optlenbuf, *opt_totlen, *optpadlen, *optleadpadlen, *opttailpadlen;
  uint8_t **buffer, **options, *optionsbuf, **optleadpad, *opttailpad;
  char *value;
  int *bufoptlen, *xtemp, *ytemp, indx;
  GtkTextBuffer *textbuffer;

  if ((io != 0) && (io != 1)) {
    fprintf (stderr, "ERROR: dst_option() received invalid value of io = %i.\n", io);
  }

  // Destination header data
  maxoptions = MAX_DSTOPTIONS;
  maxoptlen = MAX_DSTOPTLEN;
  if (data->dst_fl_flag[io]) {  // Last destination header
    nopt = &data->dstl_nopt[data->exthdr_type][io];  // (*nopt) = int
    options = data->dstl_options[data->exthdr_type][io];  // options[option #] = uint8_t *
    optlen = data->dstl_optlen[data->exthdr_type][io];  // optlen[option #] = int
    x = data->dstl_x[data->exthdr_type][io];  // x[option #] = int
    y = data->dstl_y[data->exthdr_type][io];  // y[option #] = int
    opt_totlen = &data->dstl_opt_totlen[data->exthdr_type][io];  // (*opt_totlen) = int
    optpadlen = &data->dstl_optpadlen[data->exthdr_type][io];  // (*optpadlen) = int
    optleadpadlen = data->dstl_optleadpadlen[data->exthdr_type][io];  // optleadpadlen[option #] = int
    optleadpad = data->dstl_optleadpad[data->exthdr_type][io];  // optleadpad[option #] = uint8_t *
    opttailpadlen = &data->dstl_opttailpadlen[data->exthdr_type][io];  // (*opttailpadlen) = int
    opttailpad = data->dstl_opttailpad[data->exthdr_type][io];  // opttailpad = uint8_t *
  } else {  // First destination header
    nopt = &data->dstf_nopt[data->exthdr_type][io];  // (*nopt) = int
    options = data->dstf_options[data->exthdr_type][io];  // options[option #] = uint8_t *
    optlen = data->dstf_optlen[data->exthdr_type][io];  // optlen[option #] = int
    x = data->dstf_x[data->exthdr_type][io];  // x[option #] = int
    y = data->dstf_y[data->exthdr_type][io];  // y[option #] = int
    opt_totlen = &data->dstf_opt_totlen[data->exthdr_type][io];  // (*opt_totlen) = int
    optpadlen = &data->dstf_optpadlen[data->exthdr_type][io];  // (*optpadlen) = int
    optleadpadlen = data->dstf_optleadpadlen[data->exthdr_type][io];  // optleadpadlen[option #] = int
    optleadpad = data->dstf_optleadpad[data->exthdr_type][io];  // optleadpad[option #] = uint8_t *
    opttailpadlen = &data->dstf_opttailpadlen[data->exthdr_type][io];  // (*opttailpadlen) = int
    opttailpad = data->dstf_opttailpad[data->exthdr_type][io];  // opttailpad = uint8_t *
  }
  hdrlen = DST_HDRLEN;
  optionsbuf = data->dst_optionsbuf[data->exthdr_type][io];  // optionsbuf = uint8_t *
  optlenbuf = &data->dst_optlenbuf[data->exthdr_type][io];  // (*optlenbuf) = int
  if (io == 0) {
    entryx_text = gtk_entry_get_text (GTK_ENTRY (data->dst_entry156));  // Alignment factor x
    entryy_text = gtk_entry_get_text (GTK_ENTRY (data->dst_entry157));  // Alignment factor y
    data->parent = data->dst_window;
  } else {
    entryx_text = gtk_entry_get_text (GTK_ENTRY (data->dst_entry156_outer));  // Alignment factor x
    entryy_text = gtk_entry_get_text (GTK_ENTRY (data->dst_entry157_outer));  // Alignment factor y
    data->parent = data->dst_window_outer;
  }

  // Abort if trying to insert option after non-existent option.
  if ((mode == 0) && ((opt_number > (*nopt)) || !(*nopt) || !opt_number)) {
    return (EXIT_FAILURE);
  }

  // Abort if nothing in option entry buffer to insert or append.
  if (((mode == 0) || (mode == 2)) && ((*optlenbuf) == 0)) {
    return (EXIT_FAILURE);
  }

  // Abort if trying to remove an option and none exist.
  if ((mode == 1) && ((*nopt) == 0)) {
    return (EXIT_FAILURE);
  }

  // Abort if trying to remove an option and it, specifically, doesn't exist.
  if ((mode == 1) && ((opt_number > (*nopt)) || !opt_number)) {
    return (EXIT_FAILURE);
  }

  // Abort if user requests for more than MAX_DSTOPTIONS destination options (for each of "first" and "last" destination headers).
  if (((mode == 0) || (mode == 2)) && ((*nopt) == MAX_DSTOPTIONS)) {
    sprintf (data->error_text, "The number of destination options is currently limited to %i in main.h as MAX_DSTOPTIONS.", MAX_DSTOPTIONS);
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Abort if option alignment parameters are not specified and user wants to
  // insert or append an option.
  if (((mode == 0) || (mode == 2)) && ((strnlen (entryx_text, 2) == 0) || (strnlen (entryy_text, 2) == 0))) {
    sprintf (data->error_text, "Alignment parameters x and y need to be specified.");
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Abort if alignment parameter y is greater or equal to alignment parameter x.
  if (((mode == 0) || (mode == 2)) && ((ascii_to_int64 ((char *) entryy_text) >= ascii_to_int64 ((char *) entryx_text)))) {
    sprintf (data->error_text, "Alignment parameter y cannot be greater than or equal to parameter x.");
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Allocate memory for buffer to temporarily hold current options.
  buffer = allocate_ustrmemp (maxoptions);
  for (i=0; i<maxoptions; i++) {
    buffer[i] = allocate_ustrmem (maxoptlen);
  }

  // Length of each option in buffer.
  bufoptlen = allocate_intmem (maxoptions);

  // Allocate memory for array xtemp to temporarily hold current option x-alignments.
  xtemp = allocate_intmem (maxoptions);

  // Allocate memory for array ytemp to temporarily hold current option y-alignments.
  ytemp = allocate_intmem (maxoptions);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // mode values: 
  // 0 = insert, 1 = remove
  if ((mode == 0) || (mode == 1)) {
    // Copy all current options to buffer, option lengths to bufoptlen, and
    // alignment factors to x and y.
    for (i=0; i<(*nopt); i++) {
      memcpy (buffer[i], options[i], optlen[i] * sizeof (uint8_t));
      bufoptlen[i] = optlen[i];
      xtemp[i] = x[i];
      ytemp[i] = y[i];
    }

    // Clear all current options, lengths, and alignment factors.
    for (i=0; i<maxoptions; i++) {
      memset (options[i], 0, maxoptlen * sizeof (uint8_t));
      optlen[i] = 0;
      x[i] = 0;
      y[i] = 0;
    }
  }

  // Insert mode
  if (mode == 0) {
    // Restore options, lengths, and alignment factors, but
    // stop at option after which we'll insert the new option.
    j = 0;
    for (i=0; i<opt_number; i++) {
      memcpy (options[j], buffer[i], bufoptlen[i] * sizeof (uint8_t));
      optlen[j] = bufoptlen[i];
      x[j] = xtemp[i];
      y[j] = ytemp[i];
      j++;
    }

    // New option takes length of option in buffer.
    optlen[j] = (*optlenbuf);

    // Copy contents of option buffer to option array.
    memcpy (options[j], optionsbuf, (*optlenbuf) * sizeof (uint8_t));

    // New option alignment factors are taken from entries.
    x[j] = (int) ascii_to_int64 ((char *) entryx_text);
    y[j] = (int) ascii_to_int64 ((char *) entryy_text);
    j++;

    // Copy the remaining original options, lengths, and alignment factors.
    for (i=opt_number; i<(*nopt); i++) {
      memcpy (options[j], buffer[i], bufoptlen[i] * sizeof (uint8_t));
      optlen[j] = bufoptlen[i];
      x[j] = xtemp[i];
      y[j] = ytemp[i];
      j++;
    }

    // Increment the number of options by one.
    (*nopt)++;

    // First destination header
    if (!data->dst_fl_flag[io]) {
      if (io == 0) {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25));
      } else {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25_outer));
      }
      sprintf (value, "%i", (*nopt));
      gtk_text_buffer_set_text (textbuffer, value, -1);

    // Last destination header
    } else {
      if (io == 0) {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27));
      } else {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27_outer));
      }
      sprintf (value, "%i", (*nopt));
      gtk_text_buffer_set_text (textbuffer, value, -1);
    }

  // Remove mode
  } else if (mode == 1) {
    // Restore options, lengths, and alignment factors, but skip the one we want to remove.
    j = 0;
    for (i=0; i<(*nopt); i++) {
      if ((i + 1) == opt_number) {
        continue;
      } else {
        memcpy (options[j], buffer[i], bufoptlen[i] * sizeof (uint8_t));
        optlen[j] = bufoptlen[i];
        x[j] = xtemp[i];
        y[j] = ytemp[i];
        j++;
      }
    }

    // Decrement the number of options by one.
    (*nopt)--;

    // First destination header
    if (!data->dst_fl_flag[io]) {
      if (io == 0) {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25));
      } else {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25_outer));
      }
      sprintf (value, "%i", (*nopt));
      gtk_text_buffer_set_text (textbuffer, value, -1);

    // Last destination header
    } else {
      if (io == 0) {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27));
      } else {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27_outer));
      }
      sprintf (value, "%i", (*nopt));
      gtk_text_buffer_set_text (textbuffer, value, -1);
    }

  // Append mode
  } else if (mode == 2) {
    // New option takes length of option in buffer.
    optlen[*nopt] = (*optlenbuf);

    // Copy contents of option buffer to option array.
    memcpy (options[*nopt], optionsbuf, (*optlenbuf) * sizeof (uint8_t));

    // New alignment factors are taken from entries.
    x[*nopt] = (int) ascii_to_int64 ((char *) entryx_text);
    y[*nopt] = (int) ascii_to_int64 ((char *) entryy_text);

    // Increment number of options in packet.
    (*nopt)++;

    // First destination header
    if (!data->dst_fl_flag[io]) {
      if (io == 0) {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25));
      } else {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview25_outer));
      }
      sprintf (value, "%i", (*nopt));
      gtk_text_buffer_set_text (textbuffer, value, -1);

    // Last destination header
    } else {
      if (io == 0) {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27));
      } else {
        textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->dst_textview27_outer));
      }
      sprintf (value, "%i", (*nopt));
      gtk_text_buffer_set_text (textbuffer, value, -1);
    }
  }  // End if mode == 2

  // Calculate total length of options.
  (*opt_totlen) = 0;
  for (i=0; i<(*nopt); i++) {
    (*opt_totlen) += optlen[i];
  }

  // Determine total padding needed to align and pad options (Section 4.2 of RFC 8200).
  indx = 0;  // Index is zero at start of header.

  if ((*nopt) > 0) {
    indx += hdrlen; // Account for destination header length (Next Header and Header Length fields)

    // Loop through all options.
    for (i=0; i<(*nopt); i++) {

      // Pad as needed to achieve alignment requirements for option i.
      j = indx;
      option_pad (&indx, optleadpad[i], 0, x[i], y[i]);
      optleadpadlen[i] = indx - j;

      // Length of option i
      indx += optlen[i];
    }
    // Now pad last option to next 8-byte boundary.
    j = indx;
    option_pad (&indx, opttailpad, 0, 8, 0);
    (*opttailpadlen) = indx - j;

    // Total of alignments and final padding = indx - DST_HDRLEN - total length of (non-pad) options
    (*optpadlen) = indx - hdrlen - (*opt_totlen);

    // Determine length of destination header in units of 8 bytes, excluding first 8 bytes.
    i = (indx - 8) / 8;
    if (i < 0) {
      i = 0;
    }
    if (!data->dst_fl_flag[io]) {
      data->dstfhdr[data->exthdr_type][io].hdr_len = i;  // First destination header
    } else {
      data->dstlhdr[data->exthdr_type][io].hdr_len = i;  // Last destination header
    }
  } else {
    (*opt_totlen) = 0;
    (*optpadlen) = 0;
  }

  // Free allocated memory.
  for (i=0; i<maxoptions; i++) {
    free (buffer[i]);
  }
  free (buffer);
  free (bufoptlen);
  free (value);
  free (xtemp);
  free (ytemp);

  return (EXIT_SUCCESS);
}

// View original/inner destination header options.
int
dst_opt_view (int fl, SPSData *data) {

  char *string, *title;

  // Allocate memory for various arrays.
  title = allocate_strmem (TEXT_STRINGLEN);
  string = allocate_strmem (MAX_HBHOPTIONS * MAX_HBHOPTLEN * 4);  // Arbitrarily x 4 to hold all text
  
  data->parent = data->dst_window;  // Set parent window for error messaging.

  // "First" destination header
  if (!fl) {

    // Abort if there are no options to display.
    if (!data->dstf_nopt[data->exthdr_type][0]) {
      sprintf (data->error_text, "There are no original/inner \"first\" destination options to display.");
      report_error (data);
      free (title);
      free (string);
      return (EXIT_FAILURE);

    // Original/Inner destination (first) options
    } else {

      compose_exoptstring (1, data->exthdr_type, 0, string, data->dstf_nopt[data->exthdr_type][0], data);

      sprintf (title, "Original/Inner Destination Header (first) ");

      displ_options_window (data->exthdr_type, title, string);

    }  // End if Original/Inner options exist

  // "Last" destination header
  } else {
    // Abort if there are no options to display.
    if (!data->dstl_nopt[data->exthdr_type][0]) {
      sprintf (data->error_text, "There are no original/inner \"last\" destination options to display.");
      report_error (data);
      free (title);
      return (EXIT_FAILURE);

    // Original/Inner destination (last) options
    } else {

      compose_exoptstring (2, data->exthdr_type, 0, string, data->dstl_nopt[data->exthdr_type][0], data);

      sprintf (title, "Original/Inner Destination Header (last) ");

      displ_options_window (data->exthdr_type, title, string);

    }  // End if Original/Inner options exist

  }  // End if First or Last

  // Free allocated memory.
  free (title);

  return (EXIT_SUCCESS);
}

// View outer destination header options.
int
dst_opt_view_outer (int fl, SPSData *data) {

  char *string, *title;

  // Allocate memory for various arrays.
  title = allocate_strmem (TEXT_STRINGLEN);
  string = allocate_strmem (MAX_HBHOPTIONS * MAX_HBHOPTLEN * 4);  // Arbitrarily x 4 to hold all text

  data->parent = data->dst_window_outer;  // Set parent window for error messaging.

  // "First" destination header
  if (!fl) {

    // Abort if there are no options to display.
    if (!data->dstf_nopt[data->exthdr_type][1]) {
      sprintf (data->error_text, "There are no outer \"first\" destination options to display.");
      report_error (data);
      free (title);
      free (string);
      return (EXIT_FAILURE);

    // Outer destination (first) options
    } else {

      compose_exoptstring (1, data->exthdr_type, 1, string, data->dstf_nopt[data->exthdr_type][1], data);

      sprintf (title, "Outer Destination Header (first) ");

      displ_options_window (data->exthdr_type, title, string);

    }  // End if Outer options exist

  // "Last" destination header
  } else {
    // Abort if there are no options to display.
    if (!data->dstl_nopt[data->exthdr_type][1]) {
      sprintf (data->error_text, "There are no outer \"last\" destination options to display.");
      report_error (data);
      free (title);
      return (EXIT_FAILURE);

    // Outer destination (last) options
    } else {

      compose_exoptstring (2, data->exthdr_type, 1, string, data->dstl_nopt[data->exthdr_type][1], data);

      sprintf (title, "Outer Destination Header (last) ");

      displ_options_window (data->exthdr_type, title, string);

    }  // End if Outer options exist

  }  // End if First or Last

  // Free allocated memory.
  free (title);

  return (EXIT_SUCCESS);
}
