/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Hop-by-Hop (original/inner) extension header editing window
int
hop_win (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If original/inner hop-by-hop window is already open, return with no action.
  if (data->hop_flag[0]) {
    return (EXIT_FAILURE);
  } else {
    data->hop_flag[0] = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "hop-by-hop.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->hop_window = GTK_WIDGET (gtk_builder_get_object (builder, "hop_window"));

  data->hop_radiobutton59 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "hop_radiobutton59"));  // Decimal data entry
  data->hop_radiobutton60 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "hop_radiobutton60"));  // Hexadecimal data entry

  data->hop_button142 = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button142"));  // Toggle: attach / detach a hop-by-hop header
  data->hop_textview47 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_textview47"));  // Status: attached / detached
  data->hop_entry438 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry438"));  // Hop-by-hop header length
  data->hop_entry158 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry158"));  // Hop-by-hop option data
  data->hop_entry313 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry313"));  // Option alignment parameter x (of xN + y)
  data->hop_entry314 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry314"));  // Option alignment parameter y (of xN + y)
  data->hop_button94 = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button94"));  // Insert Option After...
  data->hop_entry159 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry159"));  // Option # to insert after
  data->hop_button95 = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button95"));  // Append option
  data->hop_textview22 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_textview22"));  // Number of options in hop-by-hop header
  data->hop_button96 = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button96"));  // View hop-by-hop options
  data->hop_entry312 = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry312"));  // Option number to remove
  data->hop_button97 = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button97"));  // Remove option

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->hop_window);

  // Populate/set widgets in hop-by-hop header editing window.
  hop_show (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Populate hop_window (original/inner) with contents of selected hop-by-hop header.
int
hop_show (int type, SPSData *data) {

  char *text;
  GtkTextBuffer *textbuffer22, *textbuffer47;

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);

  // Flag indicating a hop-by-hop header
  // data->hbh_hdr_flag[type][inner=0]
  textbuffer47 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview47));
  if (data->hbh_hdr_flag[data->exthdr_type][0]) {
    gtk_text_buffer_set_text (textbuffer47, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer47, "Not Attached", -1);
  }

  // Decimal or hexadecimal option data entry
  // hop-by-hop.ui will always start with hexadecimal checked.
  // data->dec_hex_hbhopt[inner=0]
  data->dec_hex_hbhopt[0] = 1;

  // Hop-by-hop (original/inner) extension header length (header + alignments + options + padding)
  sprintf (text, "%" PRIu8, (uint8_t) data->hophdr[type][0].hdr_len);
  gtk_entry_set_text (GTK_ENTRY (data->hop_entry438), text);

  // Number of original/inner hop-by-hop options
  sprintf (text, "%i", data->hbh_nopt[type][0]);
  textbuffer22 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22));
  gtk_text_buffer_set_text (textbuffer22, text, -1);

  // Free allocated memory.
  free (text);

  return (EXIT_SUCCESS);
}

// Hop-by-Hop (outer) extension header editing window
int
hop_win_outer (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If outer hop-by-hop window is already open, return with no action.
  if (data->hop_flag[1]) {
    return (EXIT_FAILURE);
  } else {
    data->hop_flag[1] = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "hop-by-hop-outer.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->hop_window_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_window_outer"));

  data->hop_radiobutton59_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "hop_radiobutton59_outer"));  // Decimal data entry
  data->hop_radiobutton60_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "hop_radiobutton60_outer"));  // Hexadecimal data entry

  data->hop_button142_outer = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button142_outer"));  // Toggle: attach / detach a hop-by-hop header
  data->hop_textview47_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_textview47_outer"));  // Status: attached / detached
  data->hop_entry438_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry438_outer"));  // Hop-by-hop header length
  data->hop_entry158_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry158_outer"));  // Hop-by-hop option data
  data->hop_entry313_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry313_outer"));  // Option alignment parameter x (of xN + y)
  data->hop_entry314_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry314_outer"));  // Option alignment parameter y (of xN + y)
  data->hop_button94_outer = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button94_outer"));  // Insert Option After...
  data->hop_entry159_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry159_outer"));  // Option # to insert after
  data->hop_button95_outer = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button95_outer"));  // Append option
  data->hop_textview22_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_textview22_outer"));  // Number of options in hop-by-hop header
  data->hop_button96_outer = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button96_outer"));  // View hop-by-hop options
  data->hop_entry312_outer = GTK_WIDGET (gtk_builder_get_object (builder, "hop_entry312_outer"));  // Option number to remove
  data->hop_button97_outer = GTK_BUTTON (gtk_builder_get_object (builder, "hop_button97_outer"));  // Remove option

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->hop_window_outer);

  // Populate/set widgets in hop-by-hop header editing window.
  hop_show_outer (data->exthdr_type, data);

  return (EXIT_SUCCESS);
}

// Populate hop_window_outer with contents of selected outer hop-by-hop header.
int
hop_show_outer (int type, SPSData *data) {

  char *text;
  GtkTextBuffer *textbuffer22_outer, *textbuffer47_outer;

  // Array to hold various strings of text
  text = allocate_strmem (TEXT_STRINGLEN);

  // Flag indicating an outer hop-by-hop header
  // data->hbh_hdr_flag[type][outer=1]
  textbuffer47_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview47_outer));
  if (data->hbh_hdr_flag[data->exthdr_type][1]) {
    gtk_text_buffer_set_text (textbuffer47_outer, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer47_outer, "Not Attached", -1);
  }

  // Decimal or hexadecimal option data entry
  // hop-by-hop-outer.ui will always start with hexadecimal checked.
  // data->dec_hex_hbhopt[outer=1]
  data->dec_hex_hbhopt[1] = 1;

  // Hop-by-hop (outer) extension header length (header + alignments + options + padding)
  sprintf (text, "%" PRIu8, (uint8_t) data->hophdr[type][1].hdr_len);
  gtk_entry_set_text (GTK_ENTRY (data->hop_entry438_outer), text);

  // Number of outer hop-by-hop options
  sprintf (text, "%i", data->hbh_nopt[type][1]);
  textbuffer22_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22_outer));
  gtk_text_buffer_set_text (textbuffer22_outer, text, -1);

  // Free allocated memory.
  free (text);

  return (EXIT_SUCCESS);
}


// Insert, remove, or append a hop-by-hop option.
int
hop_option (int io, int mode, int opt_number, SPSData *data) {

  // Values for io:
  // 0 = original/inner hop-by-hop header
  // 1 = outer hop-by-hop-header

  // Values for mode:
  // 0 = insert
  // 1 = remove
  // 2 = append

  const char *entryx_text, *entryy_text;
  int i, j, maxoptions, maxoptlen, hdrlen;
  int *nopt, *optlen, *x, *y, *optlenbuf, *opt_totlen, *optpadlen, *optleadpadlen, *opttailpadlen;
  uint8_t **buffer, **options, *optionsbuf, **optleadpad, *opttailpad;
  char *value;
  int *bufoptlen, *xtemp, *ytemp, indx;
  GtkTextBuffer *textbuffer;

  if ((io != 0) && (io != 1)) {
    fprintf (stderr, "ERROR: hop_option() received invalid value of io = %i.\n", io);
  }

  // Hop-by-hop header data
  maxoptions = MAX_HBHOPTIONS;
  maxoptlen = MAX_HBHOPTLEN;
  nopt = &data->hbh_nopt[data->exthdr_type][io];  // (*nopt) = int
  options = data->hbh_options[data->exthdr_type][io];  // options[option #] = uint8_t *
  optlen = data->hbh_optlen[data->exthdr_type][io];  // optlen[option #] = int
  x = data->hbh_x[data->exthdr_type][io];  // x[option #] = int
  y = data->hbh_y[data->exthdr_type][io];  // y[option #] = int
  optionsbuf = data->hbh_optionsbuf[data->exthdr_type][io];  // optionsbuf = uint8_t *
  optlenbuf = &data->hbh_optlenbuf[data->exthdr_type][io];  // (*optlenbuf) = int
  opt_totlen = &data->hbh_opt_totlen[data->exthdr_type][io];  // (*opt_totlen) = int
  optpadlen = &data->hbh_optpadlen[data->exthdr_type][io];  // (*optpadlen) = int
  optleadpadlen = data->hbh_optleadpadlen[data->exthdr_type][io];  // optleadpadlen[option #] = int
  optleadpad = data->hbh_optleadpad[data->exthdr_type][io];  // optleadpad[option #] = uint8_t *
  opttailpadlen = &data->hbh_opttailpadlen[data->exthdr_type][io];  // (*opttailpadlen) = int
  opttailpad = data->hbh_opttailpad[data->exthdr_type][io];  // opttailpad = uint8_t *
  hdrlen = HOP_HDRLEN;
  if (io == 0) {
    entryx_text = gtk_entry_get_text (GTK_ENTRY (data->hop_entry313));  // Alignment factor x
    entryy_text = gtk_entry_get_text (GTK_ENTRY (data->hop_entry314));  // Alignment factor y
    data->parent = data->hop_window;
  } else {
    entryx_text = gtk_entry_get_text (GTK_ENTRY (data->hop_entry313_outer));  // Alignment factor x
    entryy_text = gtk_entry_get_text (GTK_ENTRY (data->hop_entry314_outer));  // Alignment factor y
    data->parent = data->hop_window_outer;
  }

  // Abort if trying to insert option after non-existent option.
  if ((mode == 0) && ((opt_number > (*nopt)) || !(*nopt) || !opt_number)) {
    return (EXIT_FAILURE);
  }

  // Abort if nothing in option entry buffer to insert or append.
  if (((mode == 0) || (mode == 2)) && ((*optlenbuf) == 0)) {
    return (EXIT_FAILURE);
  }

  // Abort if trying to remove an option and none exist.
  if ((mode == 1) && ((*nopt) == 0)) {
    return (EXIT_FAILURE);
  }

  // Abort if trying to remove an option and it, specifically, doesn't exist.
  if ((mode == 1) && ((opt_number > (*nopt)) || !opt_number)) {
    return (EXIT_FAILURE);
  }

  // Abort if user requests more than MAX_HBHOPTIONS hop-by-hop options.
  if (((mode == 0) || (mode == 2)) && ((*nopt) == MAX_HBHOPTIONS)) {
    sprintf (data->error_text, "The number of hop-by-hop options is currently limited to %i in main.h as MAX_HBHOPTIONS.", MAX_HBHOPTIONS);
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Abort if option alignment parameters are not specified and user wants to
  // insert or append an option.
  if (((mode == 0) || (mode == 2)) && ((strnlen (entryx_text, 2) == 0) || (strnlen (entryy_text, 2) == 0))) {
    sprintf (data->error_text, "Alignment parameters x and y need to be specified.");
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Abort if alignment parameter y is greater or equal to alignment parameter x.
  if (((mode == 0) || (mode == 2)) && ((ascii_to_int64 ((char *) entryy_text) >= ascii_to_int64 ((char *) entryx_text)))) {
    sprintf (data->error_text, "Alignment parameter y cannot be greater than or equal to parameter x.");
    report_error (data);
    return (EXIT_FAILURE);
  }

  // Allocate memory for buffer to temporarily hold current options.
  buffer = allocate_ustrmemp (maxoptions);
  for (i=0; i<maxoptions; i++) {
    buffer[i] = allocate_ustrmem (maxoptlen);
  }

  // Length of each option in buffer.
  bufoptlen = allocate_intmem (maxoptions);

  // Allocate memory for array xtemp to temporarily hold current option x-alignments.
  xtemp = allocate_intmem (maxoptions);

  // Allocate memory for array ytemp to temporarily hold current option y-alignments.
  ytemp = allocate_intmem (maxoptions);

  // String to hold miscellaneous calculated values
  value = allocate_strmem (TMP_STRINGLEN);

  // mode values: 
  // 0 = insert, 1 = remove
  if ((mode == 0) || (mode == 1)) {
    // Copy all current options to buffer, option lengths to bufoptlen, and
    // alignment factors to x and y.
    for (i=0; i<(*nopt); i++) {
      memcpy (buffer[i], options[i], optlen[i] * sizeof (uint8_t));
      bufoptlen[i] = optlen[i];
      xtemp[i] = x[i];
      ytemp[i] = y[i];
    }

    // Clear all current options, lengths, and alignment factors.
    for (i=0; i<maxoptions; i++) {
      memset (options[i], 0, maxoptlen * sizeof (uint8_t));
      optlen[i] = 0;
      x[i] = 0;
      y[i] = 0;
    }
  }

  // Insert mode
  if (mode == 0) {
    // Restore options, lengths, and alignment factors, but
    // stop at option after which we'll insert the new option.
    j = 0;
    for (i=0; i<opt_number; i++) {
      memcpy (options[j], buffer[i], bufoptlen[i] * sizeof (uint8_t));
      optlen[j] = bufoptlen[i];
      x[j] = xtemp[i];
      y[j] = ytemp[i];
      j++;
    }

    // New option takes length of option in buffer.
    optlen[j] = (*optlenbuf);

    // Copy contents of option buffer to option array.
    memcpy (options[j], optionsbuf, (*optlenbuf) * sizeof (uint8_t));

    // New option alignment factors are taken from entries.
    x[j] = (int) ascii_to_int64 ((char *) entryx_text);
    y[j] = (int) ascii_to_int64 ((char *) entryy_text);
    j++;

    // Copy the remaining original options, lengths, and alignment factors.
    for (i=opt_number; i<(*nopt); i++) {
      memcpy (options[j], buffer[i], bufoptlen[i] * sizeof (uint8_t));
      optlen[j] = bufoptlen[i];
      x[j] = xtemp[i];
      y[j] = ytemp[i];
      j++;
    }

    // Increment the number of options by one.
    (*nopt)++;

    // Update number of options in UI.
    if (io == 0) {
      textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22));
    } else {
      textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22_outer));
    }
    sprintf (value, "%i", (*nopt));
    gtk_text_buffer_set_text (textbuffer, value, -1);

  // Remove mode
  } else if (mode == 1) {
    // Restore options, lengths, and alignment factors, but skip the one we want to remove.
    j = 0;
    for (i=0; i<(*nopt); i++) {
      if ((i + 1) == opt_number) {
        continue;
      } else {
        memcpy (options[j], buffer[i], bufoptlen[i] * sizeof (uint8_t));
        optlen[j] = bufoptlen[i];
        x[j] = xtemp[i];
        y[j] = ytemp[i];
        j++;
      }
    }

    // Decrement the number of options by one.
    (*nopt)--;

    // Update number of options in UI.
    if (io == 0) {
      textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22));
    } else {
      textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22_outer));
    }
    sprintf (value, "%i", (*nopt));
    gtk_text_buffer_set_text (textbuffer, value, -1);

  // Append mode
  } else if (mode == 2) {
    // New option takes length of option in buffer.
    optlen[*nopt] = (*optlenbuf);

    // Copy contents of option buffer to option array.
    memcpy (options[*nopt], optionsbuf, (*optlenbuf) * sizeof (uint8_t));

    // New alignment factors are taken from entries.
    x[*nopt] = (int) ascii_to_int64 ((char *) entryx_text);
    y[*nopt] = (int) ascii_to_int64 ((char *) entryy_text);

    // Increment number of options in packet.
    (*nopt)++;

    // Update number of options in UI.
    if (io == 0) {
      textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22));
    } else {
      textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->hop_textview22_outer));
    }
    sprintf (value, "%i", (*nopt));
    gtk_text_buffer_set_text (textbuffer, value, -1);

  }  // End if mode == 2

  // Calculate total length of options.
  (*opt_totlen) = 0;
  for (i=0; i<(*nopt); i++) {
    (*opt_totlen) += optlen[i];
  }

  // Determine total padding needed to align and pad options (Section 4.2 of RFC 8200).
  indx = 0;  // Index is zero at start of header.

  if ((*nopt) > 0) {
    indx += hdrlen; // Account for hop-by-hop header length (Next Header and Header Length fields)

    // Loop through all options.
    for (i=0; i<(*nopt); i++) {

      // Pad as needed to achieve alignment requirements for option i.
      j = indx;
      option_pad (&indx, optleadpad[i], 0, x[i], y[i]);
      optleadpadlen[i] = indx - j;

      // Length of option i
      indx += optlen[i];
    }
    // Now pad last option to next 8-byte boundary.
    j = indx;
    option_pad (&indx, opttailpad, 0, 8, 0);
    (*opttailpadlen) = indx - j;

    // Total of alignments and final padding = indx - HOP_HDRLEN - total length of (non-pad) options
    (*optpadlen) = indx - hdrlen - (*opt_totlen);

    // Determine length of hop-by-hop header in units of 8 bytes, excluding first 8 bytes.
    i = (indx - 8) / 8;
    if (i < 0) {
      i = 0;
    }
    data->hophdr[data->exthdr_type][io].hdr_len = i;

  } else {
    (*opt_totlen) = 0;
    (*optpadlen) = 0;
  }

  // Free allocated memory.
  for (i=0; i<maxoptions; i++) {
    free (buffer[i]);
  }
  free (buffer);
  free (bufoptlen);
  free (value);
  free (xtemp);
  free (ytemp);

  return (EXIT_SUCCESS);
}

// View original/inner hop-by-hop header options.
int
hop_opt_view (SPSData *data) {

  char *string, *title;

  // Allocate memory for various arrays.
  title = allocate_strmem (TEXT_STRINGLEN);
  string = allocate_strmem (MAX_DSTOPTIONS * MAX_DSTOPTLEN * 4);  // Arbitrarily x 4 to hold all text
  
  data->parent = data->hop_window;  // Set parent window for error messaging.

  // Abort if there are no options to display.
  if (!data->hbh_nopt[data->exthdr_type][0]) {
    sprintf (data->error_text, "There are no original/inner hop-by-hop options to display.");
    report_error (data);
    free (title);
    free (string);
    return (EXIT_FAILURE);
  }

  // Original/Inner hop-by-hop options
  if (data->hbh_nopt[data->exthdr_type][0] > 0) {

    compose_exoptstring (0, data->exthdr_type, 0, string, data->hbh_nopt[data->exthdr_type][0], data);

    sprintf (title, "Original/Inner Hop-by-hop Header ");

    displ_options_window (data->exthdr_type, title, string);

  }  // End if Original/Inner options exist

  // Free allocated memory.
  free (title);
  free (string);

  return (EXIT_SUCCESS);
}

// View outer hop-by-hop header options.
int
hop_opt_view_outer (SPSData *data) {

  char *string, *title;

  // Allocate memory for various arrays.
  title = allocate_strmem (TEXT_STRINGLEN);
  string = allocate_strmem (MAX_DSTOPTIONS * MAX_DSTOPTLEN * 4);  // Arbitrarily x 4 to hold all text

  data->parent = data->hop_window_outer;  // Set parent window for error messaging.

  // Abort if there are no options to display.
  if (!data->hbh_nopt[data->exthdr_type][1]) {
    sprintf (data->error_text, "There are no outer hop-by-hop options to display.");
    report_error (data);
    free (title);
    free (string);
    return (EXIT_FAILURE);
  }

  // Outer hop-by-hop options
  if (data->hbh_nopt[data->exthdr_type][1] > 0) {

    compose_exoptstring (0, data->exthdr_type, 1, string, data->hbh_nopt[data->exthdr_type][1], data);

    sprintf (title, "Outer Hop-by-hop Header ");

    displ_options_window (data->exthdr_type, title, string);

  }  // End if Outer options exist

  // Free allocated memory.
  free (title);
  free (string);

  return (EXIT_SUCCESS);
}
