/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011-2023  Hohlraum (h0h1r4um@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

// Routing (original/inner) extension header editing window
int
route_win (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If routing window is already open, return with no action.
  if (data->route_flag[0]) {
    return (EXIT_FAILURE);
  } else {
    data->route_flag[0] = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "routing.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->route_window = GTK_WIDGET (gtk_builder_get_object (builder, "route_window"));

  data->rt_button145 = GTK_BUTTON (gtk_builder_get_object (builder, "rt_button145"));  // Toggle: attach / detach a routing header
  data->rt_textview50 = GTK_WIDGET (gtk_builder_get_object (builder, "rt_textview50"));  // Status: attached / detached
  data->rt_entry446 = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry446"));  // Header length
  data->rt_entry447 = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry447"));  // Routing type
  data->rt_entry448 = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry448"));  // Segments left
  data->rt_radiobutton73 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "rt_radiobutton73"));  // Decimal routing data entry
  data->rt_radiobutton74 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "rt_radiobutton74"));  // Hexadecimal routing data entry
  data->rt_entry449 = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry449"));  // Routing data
  data->rt_button125 = GTK_BUTTON (gtk_builder_get_object (builder, "rt_button125"));  // Load ICV from file...

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->route_window);

  // Populate/set widgets in routing header editing window.
  route_show (data);

  return (EXIT_SUCCESS);
}

// Populate route_window with contents of selected routing header.
int
route_show (SPSData *data) {

  char *value;
  GtkTextBuffer *textbuffer50;

  // Strings to hold miscellaneous calculated values
  value = allocate_strmem (IP_MAXPACKET);

  // Flag indicating a routing header
  textbuffer50 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->rt_textview50));
  if (data->route_hdr_flag[data->exthdr_type][0]) {
    gtk_text_buffer_set_text (textbuffer50, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer50, "Not Attached", -1);
  }

  // Decimal or hexadecimal routing data entry
  // routing.ui will always start with hexadecimal checked.
  data->dec_hex_route[0] = 1;

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Routing (outer) extension header editing window
int
route_win_outer (SPSData *data) {

  GtkBuilder *builder;
  GError *error = NULL;

  // If routing window is already open, return with no action.
  if (data->route_flag[1]) {
    return (EXIT_FAILURE);
  } else {
    data->route_flag[1] = 1;
  }

  // Create new GtkBuilder object.
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "routing-outer.ui", &error)) {
    g_warning ("%s", error->message);
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Get objects from UI file
  data->route_window_outer = GTK_WIDGET (gtk_builder_get_object (builder, "route_window_outer"));

  data->rt_button145_outer = GTK_BUTTON (gtk_builder_get_object (builder, "rt_button145_outer"));  // Toggle: attach / detach a routing header
  data->rt_textview50_outer = GTK_WIDGET (gtk_builder_get_object (builder, "rt_textview50_outer"));  // Status: attached / detached
  data->rt_entry446_outer = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry446_outer"));  // Header length
  data->rt_entry447_outer = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry447_outer"));  // Routing type
  data->rt_entry448_outer = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry448_outer"));  // Segments left
  data->rt_radiobutton73_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "rt_radiobutton73_outer"));  // Decimal routing data entry
  data->rt_radiobutton74_outer = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "rt_radiobutton74_outer"));  // Hexadecimal routing data entry
  data->rt_entry449_outer = GTK_WIDGET (gtk_builder_get_object (builder, "rt_entry449_outer"));  // Routing data
  data->rt_button125_outer = GTK_BUTTON (gtk_builder_get_object (builder, "rt_button125_outer"));  // Load ICV from file...

  // Connect signals.
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore.
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder.
  gtk_widget_show (data->route_window_outer);

  // Populate/set widgets in routing header editing window.
  route_show_outer (data);

  return (EXIT_SUCCESS);
}

// Populate route_window_outer with contents of selected routing header.
int
route_show_outer (SPSData *data) {

  char *value;
  GtkTextBuffer *textbuffer50_outer;

  // Strings to hold miscellaneous calculated values
  value = allocate_strmem (IP_MAXPACKET);

  // Flag indicating a routing header
  textbuffer50_outer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->rt_textview50_outer));
  if (data->route_hdr_flag[data->exthdr_type][1]) {
    gtk_text_buffer_set_text (textbuffer50_outer, "Attached", -1);
  } else {
    gtk_text_buffer_set_text (textbuffer50_outer, "Not Attached", -1);
  }

  // Decimal or hexadecimal routing data entry
  // routing.ui will always start with hexadecimal checked.
  data->dec_hex_route[1] = 1;

  // Free allocated memory.
  free (value);

  return (EXIT_SUCCESS);
}

// Update routing header length.
int
update_route_len (int type, int io, SPSData *data) {

  // Length of routing header (8-byte units), excluding first 8 bytes
  // Section 4.4 of RFC 8200
  // If routing data is >= 4 bytes, then we have 8+ in total; RTE_HDRLEN = 4, so subtract 4 from data, for a total subtraction of 8.
  if (data->route_datlen[type][io] >= 4 ) {
    data->routehdr[type][io].hdr_len = ((data->route_datlen[type][io] - 4u) / 8u);  // Don't count 1st 4 bytes of routing data.
  } else {
    data->routehdr[type][io].hdr_len = 0u;
  }

  return (EXIT_SUCCESS);
}
